=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2007-04-20 04:18:37 +0000
+++ bzrlib/fetch.py	2007-07-02 05:05:25 +0000
@@ -132,18 +132,33 @@
         try:
             pp.next_phase()
             revs = self._revids_to_fetch()
-            # something to do ?
-            if revs:
-                pp.next_phase()
-                self._fetch_weave_texts(revs)
-                pp.next_phase()
+            self._fetch_everything_for_revisions(revs, pp)
+        finally:
+            self.pb.clear()
+
+    def _fetch_everything_for_revisions(self, revs, pp):
+        """Fetch all data for the given set of revisions."""
+        if revs is None:
+            return
+        what_to_do = self.from_repository.get_data_about_revision_ids(revs)
+        for knit_kind, file_id, revisions in what_to_do:
+            if knit_kind == "file":
+                self._fetch_weave_text(file_id, revisions)
+            elif knit_kind == "inventory":
+                # XXX:
+                # Once we've processed all the files, then we generate the root
+                # texts (if necessary), then we process the inventory.  It's a
+                # bit distasteful to have knit_kind == "inventory" mean this,
+                # perhaps it should happen on the first non-"file" knit, in case
+                # it's not always inventory?
+                self._generate_root_texts(revs)
                 self._fetch_inventory_weave(revs)
-                pp.next_phase()
+            elif knit_kind == "revisions":
                 self._fetch_revision_texts(revs)
-                self.count_copied += len(revs)
-        finally:
-            self.pb.clear()
-
+            else:
+                raise AssertionError("Unknown knit kind %r" % knit_kind)
+        self.count_copied += len(revs)
+        
     def _revids_to_fetch(self):
         mutter('fetch up to rev {%s}', self._last_revision)
         if self._last_revision is NULL_REVISION:
@@ -173,25 +188,28 @@
             for file_id, required_versions in file_ids.items():
                 texts_pb.update("fetch texts", count, num_file_ids)
                 count +=1
-                to_weave = self.to_weaves.get_weave_or_empty(file_id,
-                    self.to_repository.get_transaction())
-                from_weave = self.from_weaves.get_weave(file_id,
-                    self.from_repository.get_transaction())
-                # we fetch all the texts, because texts do
-                # not reference anything, and its cheap enough
-                to_weave.join(from_weave, version_ids=required_versions)
-                # we don't need *all* of this data anymore, but we dont know
-                # what we do. This cache clearing will result in a new read 
-                # of the knit data when we do the checkout, but probably we
-                # want to emit the needed data on the fly rather than at the
-                # end anyhow.
-                # the from weave should know not to cache data being joined,
-                # but its ok to ask it to clear.
-                from_weave.clear_cache()
-                to_weave.clear_cache()
+                self._fetch_weave_text(file_id, required_versions)
         finally:
             texts_pb.finished()
 
+    def _fetch_weave_text(self, file_id, required_versions):
+        to_weave = self.to_weaves.get_weave_or_empty(file_id,
+            self.to_repository.get_transaction())
+        from_weave = self.from_weaves.get_weave(file_id,
+            self.from_repository.get_transaction())
+        # we fetch all the texts, because texts do
+        # not reference anything, and its cheap enough
+        to_weave.join(from_weave, version_ids=required_versions)
+        # we don't need *all* of this data anymore, but we dont know
+        # what we do. This cache clearing will result in a new read 
+        # of the knit data when we do the checkout, but probably we
+        # want to emit the needed data on the fly rather than at the
+        # end anyhow.
+        # the from weave should know not to cache data being joined,
+        # but its ok to ask it to clear.
+        from_weave.clear_cache()
+        to_weave.clear_cache()
+
     def _fetch_inventory_weave(self, revs):
         pb = bzrlib.ui.ui_factory.nested_progress_bar()
         try:
@@ -218,6 +236,15 @@
         finally:
             pb.finished()
 
+    def _generate_root_texts(self, revs):
+        """This will be called by __fetch between fetching weave texts and
+        fetching the inventory weave.
+
+        Subclasses should override this if they need to generate root texts
+        after fetching weave texts.
+        """
+        pass
+        
 
 class GenericRepoFetcher(RepoFetcher):
     """This is a generic repo to repo fetcher.
@@ -359,6 +386,9 @@
 
     def _fetch_weave_texts(self, revs):
         GenericRepoFetcher._fetch_weave_texts(self, revs)
+        self._generate_root_texts(revs)
+
+    def _generate_root_texts(self, revs):
         # Now generate a weave for the tree root
         self.helper.generate_root_texts(revs)
 
@@ -377,6 +407,9 @@
 
     def _fetch_weave_texts(self, revs):
         KnitRepoFetcher._fetch_weave_texts(self, revs)
+        self._generate_root_texts(revs)
+
+    def _generate_root_texts(self, revs):
         # Now generate a weave for the tree root
         self.helper.generate_root_texts(revs)
 

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2007-06-09 01:13:31 +0000
+++ bzrlib/repository.py	2007-07-02 05:05:25 +0000
@@ -616,6 +616,35 @@
             pb.finished()
         return result
 
+    def get_data_about_revision_ids(self, revision_ids):
+        """Get an iterable about data for a given set of revision IDs.
+
+        The named data will be ordered so that it can be fetched and inserted in
+        that order safely.
+        
+        :returns: (knit-kind, file-id, versions)
+        """
+        # XXX: it's a bit weird to control the inventory weave caching in this
+        # generator.  Ideally the caching would be done in fetch.py I think.
+        inv_w = self.get_inventory_weave()
+        inv_w.enable_cache()
+
+        # file ids that changed
+        file_ids = self.fileids_altered_by_revision_ids(revision_ids)
+        for file_id, altered_versions in file_ids.iteritems():
+            yield ("file", file_id, altered_versions)
+
+        # inventory
+        yield ("inventory", None, revision_ids)
+        inv_w.clear_cache()
+
+        # signatures
+        # XXX
+
+        # revisions
+        yield ("revisions", None, revision_ids)
+
+
     @needs_read_lock
     def get_inventory_weave(self):
         return self.control_weaves.get_weave('inventory',

